/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   SMB "NET TIME /SET" utility
   Copyright (C) Andrew Tridgell 1994-1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define NO_SYSLOG

#include "includes.h"

#ifndef REGISTER
#define REGISTER 0
#endif

struct cli_state *cli;
extern BOOL in_client;
static int port = SMB_PORT;
static pstring service;
extern pstring global_myname;
static pstring username;
static pstring workgroup;
extern struct in_addr ipzero;
extern pstring scope;

static int name_type = 0x20;

extern pstring user_socket_options;

extern pstring debugf;
extern int DEBUGLEVEL;

BOOL translation = False;

static BOOL have_ip;

struct in_addr dest_ip;

/****************************************************************************
usage on the program
****************************************************************************/
static void usage(char *pname)
{
  DEBUG(0,("Usage: %s server [options]", pname));

  DEBUG(0,("\nVersion %s\n",VERSION));
  DEBUG(0,("\t-s smb.conf           pathname to smb.conf file\n"));
  DEBUG(0,("\t-O socket_options     socket options to use\n"));
  DEBUG(0,("\t-R name resolve order use these name resolution services
only\n"));
  DEBUG(0,("\t-i scope              use this NetBIOS scope\n"));
  DEBUG(0,("\t-n netbios name.      Use this name as my netbios name\n"));
  DEBUG(0,("\t-d debuglevel         set the debuglevel\n"));
  DEBUG(0,("\t-p port               connect to the specified port\n"));
  DEBUG(0,("\t-l log basename.      Basename for log/debug files\n"));
  DEBUG(0,("\t-h                    Print this help message.\n"));
  DEBUG(0,("\t-I dest IP            use this IP to connect to\n"));
  DEBUG(0,("\t-E                    write messages to stderr instead of
stdout\n"));
  DEBUG(0,("\t-m max protocol       set the max protocol level\n"));
  DEBUG(0,("\t-W workgroup          set the workgroup name\n"));
  DEBUG(0,("\n"));
}

/*
  normalize_tv places tv_usec in the range 0 ... 999999
  when it is in the range -999999 ... 1999999 i.e. after an addition
  or subtraction of normalized values.
*/

static void normalize_tv(struct timeval *tv)
{
	if(tv->tv_usec < 0)
		tv->tv_sec--, tv->tv_usec += 1000000;
	else if(tv->tv_usec >= 1000000)
		tv->tv_sec++, tv->tv_usec -= 1000000;
}

/* Perform tv1 -= tv2 */
void subtract_tv(struct timeval *tv1,
				 struct timeval *tv2)
{
	tv1->tv_sec -= tv2->tv_sec;
	tv1->tv_usec -= tv2->tv_usec;
	normalize_tv(tv1);
}

/***************************************************** 
get the server's time and set ourselves to match.
*******************************************************/
int do_nettime(char *a_server)
{
	struct cli_state *c;
	struct nmb_name called, calling;
	struct in_addr ip;
	extern struct in_addr ipzero;
	struct timeval start, end;
	int ok;
	char *server = strdup(a_server + 2); /* server name without
	
leading double \'es */

	/* Chop off share name from a_server */
	{
		char *sp = strchr(server, '\\');
		if(sp)
			*sp = 0;
	}

	ip = ipzero;

	make_nmb_name(&calling, global_myname, 0x0);
	make_nmb_name(&called , server, name_type);

 again:
	ip = ipzero;
	if (have_ip) ip = dest_ip;

	/* have to open a new connection */
	if (!(c=cli_initialise(NULL)) || (cli_set_port(c, port) == 0) ||
	    !cli_connect(c, server, &ip)) {
		DEBUG(0,("Connection to %s failed\n", server));
		if(c)
			cli_shutdown(c);
		free(server);
		return 0;				/* Failed to set time.
*/
	}

	if (!cli_session_request(c, &calling, &called)) {
		char *p;
		DEBUG(0,("session request to %s failed (%s)\n", 
			 called.name, cli_errstr(c)));
		cli_shutdown(c);
		if ((p=strchr(called.name, '.'))) {
			*p = 0;
			goto again;
		}
		if (strcmp(called.name, "*SMBSERVER")) {
			make_nmb_name(&called , "*SMBSERVER", 0x20);
			goto again;
		}
		free(server);
		return 0;				/* Failed to set time.
*/
	}

	gettimeofday(&start, NULL);	/* Get a precise start time */
	/* measure the time it takes to negotiate - during which time the
	   server returns it's time */

	if (!cli_negprot(c)) {
		DEBUG(0,("protocol negotiation failed\n"));
		cli_shutdown(c);
		free(server);
		return 0;
	}

	gettimeofday(&end, NULL);	/* Get a precise end time */

	ok = 0;						/* assume the time
can't be set */

	subtract_tv(&end, &start);	/* Compute duration in end */
	if(end.tv_sec == 0 &&
	   end.tv_usec <= 500000)	/* Under half a second? */
	{
		end.tv_usec = 0;
		end.tv_sec = c->servertime;
		if(settimeofday(&end, NULL))
			perror("settimeofday");
		else
		{
			time_t servertime = c->servertime;
			struct tm *tm = localtime(&servertime);
			/* Got it!  Set the system time. */
			printf("Current system time set to %s", asctime(tm));
			ok = 1;					/* Okay, system
time is set. */
		}
	}

	cli_shutdown(c);
	free(server);
	return ok;
}

/****************************************************************************
  main program
****************************************************************************/
int main(int argc,char *argv[])
{
	char *pname = argv[0];
	int opt;
	extern FILE *dbf;
	extern char *optarg;
	extern int optind;
	static pstring servicesf = CONFIGFILE;
	pstring new_name_resolve_order;

	*new_name_resolve_order = 0;

	DEBUGLEVEL = 2;

	setup_logging(pname,True);

	TimeInit();
	charset_initialise();

	in_client = True;   /* Make sure that we tell lp_load we are */

	if (!lp_load(servicesf,True,False,False)) {
		fprintf(stderr, "Can't load %s - run testparm to debug it\n",
servicesf);
	}
	
#ifdef WITH_SSL
	sslutil_init(0);
#endif

	pstrcpy(workgroup,lp_workgroup());

	load_interfaces();

	pstrcpy(username,"GUEST");

	if (argc < 2) {
		usage(pname);
		exit(1);
	}
  
	pstrcpy(service,argv[1]);  
	/* Convert any '/' characters in the service name to '\' characters */
	string_replace( service, '/','\\');
	argc--;
	argv++;
		
	if (service[0] != '\\' ||
		service[1] != '\\' ||
		count_chars(service,'\\') != 2)
	{
		usage(pname);
		printf("\n%s: improper server name %s\n", argv[0], service);
		exit(1);
	}

	while ((opt = 
		getopt(argc, argv,"s:O:R:M:i:Nn:d:Pp:l:hI:EU:t:m:W:T:D:c:b:"))
!= EOF) {
		switch (opt) {
		case 's':
			pstrcpy(servicesf, optarg);
			break;
		case 'O':
			pstrcpy(user_socket_options,optarg);
			break;	
		case 'R':
			pstrcpy(new_name_resolve_order, optarg);
			break;
		case 'i':
			pstrcpy(scope,optarg);
			break;
		case 'n':
			pstrcpy(global_myname,optarg);
			break;
		case 'd':
			if (*optarg == 'A')
				DEBUGLEVEL = 10000;
			else
				DEBUGLEVEL = atoi(optarg);
			break;
		case 'p':
			port = atoi(optarg);
			break;
		case 'l':
			slprintf(debugf,sizeof(debugf)-1, "%s.client",optarg);
			break;
		case 'h':
			usage(pname);
			exit(0);
			break;
		case 'I':
			{
				dest_ip = *interpret_addr2(optarg);
				if (is_zero_ip(dest_ip))
					exit(1);
				have_ip = True;
			}
			break;
		case 'E':
			dbf = stderr;
			break;
		case 'W':
			pstrcpy(workgroup,optarg);
			break;
		default:
			usage(pname);
			exit(1);
		}
	}

	get_myname((*global_myname)?NULL:global_myname);  

	if(*new_name_resolve_order)
		lp_set_name_resolve_order(new_name_resolve_order);

	DEBUG( 3, ( "Client started (version %s).\n", VERSION ) );

	do_nettime(service);
	return 0;
}
